//===========================================================
//=     University of Illinois at Urbana/Champaign          =
//=     Department of Computer Science                      =
//=     Dr. Dan Roth - Cognitive Computation Group          =
//=                                                         =
//=  Project: Snow                                          =
//=                                                         =
//=   Module: SnowParam.cpp                                 =
//=  Version: 2.0                                           =
//=  Authors: Jeff Rosen and Andrew Carlson                 =
//=     Date: xx/xx/99                                      = 
//=                                                         =
//= Comments:                                               =
//===========================================================

#include "GlobalParams.h"
#include "SnowParam.h"
#include <stdlib.h>
#include <iostream>
#include <fstream>

#ifdef WIN32
extern "C"
{
    extern int getopt(int, char* const *, const char*);
    extern char *optarg;
}
#endif

extern GlobalParams globalParams;

bool ProcessParam( int param, const char* arg )
{
    switch (param)
    {
    case 'B':
        // Add a naive bayes algorithm to the network
        globalParams.algorithmSpecification += "b(";
        globalParams.algorithmSpecification += arg;
        globalParams.algorithmSpecification += "),";
        break;

    case 'b':
        globalParams.bayesSmoothing = atof(optarg);
        break;

    case 'c':
        // interval for learning curve
        globalParams.curveInterval = atoi(arg);
        break;

    case 'd':
        // discarding
        if (!strcmp(arg, "none"))
        {
            globalParams.discardMethod = DISCARD_NONE;
        } else if (!strncmp(arg, "abs", 3)) {
            globalParams.discardMethod = DISCARD_ABS;
            if (arg[3] == ':')
            {
                globalParams.discardThreshold = atof(&arg[4]);
            } else {
                cerr << "ERROR: no discard threshold given for absolute discard mode" 
                     << endl;
                return false;
            } 
        } else if (!strncmp(arg, "rel", 3)) {
            globalParams.discardMethod = DISCARD_REL;
        } else {
            cerr << "ERROR: discard flag -d requires parameter abs:# or rel" << endl;
            return false;
        }
        break;

    case 'E':
        // Error report file
        globalParams.errorFile = arg;    
        break;

    case 'e':
        globalParams.eligibilityThreshold = atoi(arg);
        break;

    case 'F':
        // Network file
        globalParams.networkFile = arg;
        break;

    case 'I':
        // Input file
        globalParams.inputFile = arg;
        break;
 
    case 'i':
        if (*arg == '+')
        {
            globalParams.onlineLearning = true;
        } else if (*arg == '-')
        {
            globalParams.onlineLearning = false;
        } else {
            cerr << "ERROR: Unrecognized argument '" << arg
                 << "' to '-o' online flag" << endl;
            return false;
        }
        break;
        
    case 'l':
        if (*arg == '+')
        {
            globalParams.labelsPresent = true;
        } else if (*arg == '-')
        {
            globalParams.labelsPresent = false;
        } else {
            cerr << "ERROR: Unrecognized argument '" << arg
                 << "' to '-l' label flag" << endl;
            return false;
        }
        break;

    case 'm':
        if (*arg == '+')
        {
            globalParams.multipleLabels = true;
        } else if (*arg == '-')
        {
            globalParams.multipleLabels = false;
        } else {
            cerr << "ERROR: Unrecognized argument '" << arg
                 << "' to '-m' multiple labels flag" << endl;
            return false;
        }
        break;


    case 'o':   
        if (strcmp(arg, "accuracy") == 0) {
            globalParams.predictMethod = ACCURACY;
        } else if (strcmp(arg, "winners") == 0) {
            globalParams.predictMethod = WINNERS;
        } else if (strcmp(arg, "allactivations") == 0) {
            globalParams.predictMethod = ALL_ACTIVATIONS;
        } else if (strcmp(arg, "allpredictions") == 0) {
            globalParams.predictMethod = ALL_PREDICTIONS;
        } else if (strcmp(arg, "allboth") == 0) {
            globalParams.predictMethod = ALL_BOTH;
        } else {
            cerr << "ERROR: Unrecognized argument '" << arg
                 << "' to '-o' output mode flag" << endl;
            return false;
        }
        break;
        
    case 'P':
        // Add a perceptron algorithm to the network
        globalParams.algorithmSpecification += "p(";
        globalParams.algorithmSpecification += arg;
        globalParams.algorithmSpecification += "),";
        break;
 
    case 'p':
        globalParams.predictionThreshold = atof(arg);
        break;
    
    case 'r':
        // number of learning cycles (repetitions)
        globalParams.cycles = atoi(arg);
        break;

    case 'R':
        // results file for Test mode
        globalParams.outputFile = arg;
        break;
        
    case 's':
        // sparse network
        if (*arg == 's')
        {
            globalParams.sparseNetwork = true;
        } else if (*arg == 'f')
        {
            globalParams.sparseNetwork = false;
        } else {
            cerr << "ERROR: Unrecognized argument '" << arg
                 << "' to '-s' sparse network flag" << endl;
            return false;
        }
        break;
            
    case 'T':
        // testing file for Train/Test mode
        globalParams.testFile = arg;
        break;

   
    case 'u':
        if (*arg == '+')
        {
            globalParams.noFirstCycleUpdate = false;
        } else if (*arg == '-')
        {
            globalParams.noFirstCycleUpdate = true;
        } else {
            cerr << "ERROR: Unrecognized argument '" << arg
                 << "' to '-u' update during first cycle flag" << endl;
            return false;
        }
        break;
            
    case 'v':
        // Verbosity level
        if (!strcmp(arg, "off"))
        {
           globalParams.verbosity = VERBOSE_QUIET;
        } else if (!strcmp(arg, "med")) {
            globalParams.verbosity = VERBOSE_MED;
        } else if (!strcmp(arg, "max")) {
            globalParams.verbosity = VERBOSE_MAX;
        } else {
            globalParams.verbosity = VERBOSE_MIN;
        }
        break;

    case 'W':
        // Add a winnow algorithm to the network
        globalParams.algorithmSpecification += "w(";
        globalParams.algorithmSpecification += arg;
        globalParams.algorithmSpecification += "),";                
        break;

    case 'w':
        // Set the winnow/perceptron smoothing
        globalParams.smoothing = atof(arg);
        break;

    case 'x':
        // Example
        globalParams.evalExample = arg;
        break;
            
    default:
        cerr << "ERROR: Unrecognized command parameter" << endl;
        return false;
    }

    return true;
}

bool ParseParamFile( const char* paramFilename )
{
    bool result = true;
    char paramLine[256];
    char *arg;
    
    ifstream paramStream(paramFilename);
    if (!paramStream)
    {
        cerr << "Error:" << endl;
        cerr << "Failed to open architecture file: '"
             << paramFilename << "'. No parameters read." << endl;
        return false;
    }

    while (result && !paramStream.eof())
    {
        //string paramLine;
        paramLine[0] = 0;
        paramStream.getline(paramLine, sizeof paramLine);
        if (strlen(paramLine))
        {
            if (paramLine[0] == '-')
            {
                arg = paramLine + 2;
                while (isspace(*arg)) arg++;
                result = ProcessParam(paramLine[1], arg);
            } else if (paramLine[0] != '#')
            {
                cerr << "Error:" << endl;
                cerr << "Each line in the pararmeter file must start with '-' "
                     << "skipping " << paramLine << endl;
            }
        }
        
    }
    
    return result;
}

bool ValidateParams()
{
    bool result = true;
    
    // Always must be a network file
    if (!globalParams.networkFile.length())
    {
        cerr << "Fatal Error:" << endl;
        cerr << "No network file specified (-F)" << endl << endl;
        result = false;
    }

    // If run mode is TRAIN or TEST, need an input file 
    if ((globalParams.runMode != MODE_EVAL) &&
        !globalParams.inputFile.length())
    {
        cerr << "Fatal Error:" << endl;
        cerr << "No input file specified (-I)" << endl << endl;
        result = false;
    }
    // In TRAIN mode, check for architecture definition
    if ((globalParams.runMode == MODE_TRAIN) &&
        globalParams.targetIds.empty() &&
        !globalParams.algorithmSpecification.length())
    {
        globalParams.algorithmSpecification += "w(1.3,0.85,1.0,0.2:0-1),";
        cerr << "Warning: "  
             << "No architecture defined (defaulting to w(1.3,0.85,1.0,0.2:0-1))"
           << endl;
    }
    // If run mode is Eval, need an example
    if ((globalParams.runMode == MODE_EVAL) &&
        !globalParams.evalExample.length())
    {
        cerr << "Fatal Error:" << endl;
        cerr << "No example specified (-x)" << endl << endl;
        result = false;
    }

    // If run mode is not Eval, no example needed (used)
    if ((globalParams.runMode != MODE_EVAL) &&
        globalParams.evalExample.length())
    {
        cerr << "Warning:" << endl;
        if (globalParams.runMode == MODE_TRAIN)
        {
            cerr << "Training mode: ";
        } else {
            cerr << "Test mode: ";
        }
        cerr << "Specified example will be ignored (-x)" << endl << endl;
    }

    // If run mode is Train and online learning is specified
    if ((globalParams.runMode == MODE_TRAIN) &&
        globalParams.onlineLearning && (globalParams.testFile.length() == 0))
    {
        cerr << "Warning: Online learning doesn't make sense in" << endl
             << "training mode. Online flag will be ignored." << endl;
    }
    
    return result;
}

bool ParseCmdLine( int argc, char* argv[] )
{
    bool result = true;
    bool cmdLineNetworks = false;
    
    // Must supply either -eval, -test or -train as first argument
    if (argc > 1)
    {
        if (strcmp(argv[1], "-test") == 0)
        {
            globalParams.runMode = MODE_TEST;
        } else if (strncmp(argv[1], "-eval", 4) == 0)
        {
            globalParams.runMode = MODE_EVAL;
        } else if (strcmp(argv[1], "-train") == 0)
        {
            globalParams.runMode = MODE_TRAIN;
        } else { 
            cerr << "ERROR: First argument must be -train or -evaluate"
                 << endl << endl;
            return false;
        }
    } else { 
        cerr << "First argument should be -train or -test\n" ;
        return false;
    }

    // command line arguments

    // skip first argument (mode itself)
    ++argv;
    --argc;

    // look for parameter file first
    for (int i = 0; i < argc; i++)
    {
        if (!strncmp(argv[i], "-A", 2))
        {
            // Process the parameter file
            if (strlen(argv[i]) > 2)
            {
                result = ParseParamFile(&argv[i][2]);
            } else {
                result = ParseParamFile(argv[i+1]);
            }
        }
    }
    
    // Read in input data
    int c;
    while (((c = getopt(argc,argv,"A:b:B:c:d:e:E:F:i:I:l:L:m:o:p:P:r:R:s:T:v:w:W:x:")) != -1)
           && (result == true))
    {
        // Ignore -A, it has already been processed above
        if (c != 'A')
        {
            // If we get a network specification, let it override
            // whatever may have been in the param file
            if (!cmdLineNetworks
                && ((c == 'B') || (c == 'P') || (c == 'W')))
            {
                globalParams.algorithmSpecification.empty();
                cmdLineNetworks = true;
            }
            result = ProcessParam(c, optarg);
        }    
    }

    return result;
}
